
// ------------------------------------------------------------------------------------------------------------------------------------------------>  
// SETLIGHTS - the main function which assigns the appropriate setting to each light based on the current actual drive mode (different from commanded drive mode)
// ------------------------------------------------------------------------------------------------------------------------------------------------>  
void SetLights(int DriveMode)
{
    int SaveSetting[NumLights];

    int i;
    int j;

    // Loop through each light, assign the setting appropriate to its state
    for (j=0; j<NumLights; j++)
    {
        // We will use the temporary variable SaveSetting to assign the setting for this light. 
        // A light could have multiple settings apply at one time, be we can only set it to one thing.
        // Therefore each setting is ranked by importance. If multiple settings apply to a light, 
        // the setting applied LAST will be the one used (each check can overwrite the prior one). 
        // You can re-order the checks below, the least important should come first, and most important last.


        // Least important - does this light have a setting related to Channel 3? 
        // --------------------------------------------------------------------------------------------------->>
        SaveSetting[j] = LightSettings[j][Channel3];


        // Next - does this light have a setting related to Drive Mode? (Forward, reverse, stop)
        // --------------------------------------------------------------------------------------------------->>
        switch (DriveMode) {
            case FWD:
                if (LightSettings[j][StateFwd]  != NA) { SaveSetting[j] = LightSettings[j][StateFwd]; }
                break;
            case REV:
                if (LightSettings[j][StateRev]  != NA) { SaveSetting[j] = LightSettings[j][StateRev]; }
                break;
            case STOP:
                if (LightSettings[j][StateStop] != NA) { SaveSetting[j] = LightSettings[j][StateStop]; }
                break;
        }


        // Next - does this light come on during braking?
        // --------------------------------------------------------------------------------------------------->>        
        if (Braking)
        {
            if (LightSettings[j][StateBrake] != NA) { SaveSetting[j] = LightSettings[j][StateBrake]; }
        }
    

        // Next - does this light come on during turns? 
        // --------------------------------------------------------------------------------------------------->>        
        if (TurnCommand > 0 )    // Right Turn
        {
            // If we have a blink command on right turn, and if we have the BlinkTurnOnlyAtStop = true, and if we are stopped, and if the turn signal delay has expired (TurnSignal_Enable = true), 
            // then and only then do we apply the blink command
            if ((LightSettings[j][StateRT] == BLINK) && (BlinkTurnOnlyAtStop == true) && (DriveMode == STOP) && (TurnSignal_Enable == true))
            {
                SaveSetting[j] = BLINK;
            }
            // Otherwise if it is any other setting, or if the BlinkTurnOnlyAtStop flag is not true, then we apply the setting normally
            else if (LightSettings[j][StateRT] != NA) { SaveSetting[j] = LightSettings[j][StateRT]; }
        }
    
        if (TurnCommand < 0 )    // Left Turn
        {
            // If we have a blink command on right turn, and if we have the BlinkTurnOnlyAtStop = true, and if we are stopped, and if the turn signal delay has expired (TurnSignal_Enable = true), 
            // then and only then do we apply the blink command
            if ((LightSettings[j][StateLT] == BLINK) && (BlinkTurnOnlyAtStop == true) && (DriveMode == STOP) && (TurnSignal_Enable == true))
            {
                SaveSetting[j] = BLINK;
            }
            // Otherwise if it is any other setting, or if the BlinkTurnOnlyAtStop flag is not true, then we apply the setting normally
            else if (LightSettings[j][StateLT] != NA) { SaveSetting[j] = LightSettings[j][StateLT]; }
        }

        // Light "j" now has a single setting = SaveSetting[j]
        // We call the function that will set this light to that setting
        SetLight(j, SaveSetting[j]);
    }

    // Now we have done looping through all the lights, and setting them all. 

    // There is an option in UserConfig.h to have the board LEDs reflect car movement, this 
    // can sometimes be useful for debugging. 

    // If moving forward, green led is on, if moving backwards, red led is on. 
    // If stopped, both are off.
    // If braking, both are on. 
    if (LED_DEBUG)
    {
        if (DriveMode == FWD)  { digitalWrite(RedLED , LOW ) ; digitalWrite(GreenLED, HIGH) ; }
        if (DriveMode == REV)  { digitalWrite(RedLED , HIGH) ; digitalWrite(GreenLED, LOW ) ; }
        if (DriveMode == STOP) { digitalWrite(RedLED , LOW ) ; digitalWrite(GreenLED, LOW ) ; }    
        if (Braking == true)   { digitalWrite(RedLED , HIGH) ; digitalWrite(GreenLED, HIGH) ; }
    }

}



// ------------------------------------------------------------------------------------------------------------------------------------------------>  
// SETUPLIGHTS - This assigns the various settings to each of the light states
// ------------------------------------------------------------------------------------------------------------------------------------------------>  
void SetupLights(int WhatScheme)
{
    int i;
    int j;

    for (i=0; i<NumLights; i++)
    {
        for (j=0; j<NumStates; j++)
        {
            LightSettings[i][j] = pgm_read_word_near(&(Schemes[WhatScheme-1][i][j]));    // WhatScheme is minus -1 because Schemes are zero-based. We let the user use
        }                                                                                // one-based numbers for convenience
    }
    return;
}


// ------------------------------------------------------------------------------------------------------------------------------------------------>  
// SETUPLIGHTS - This assigns the various settings to each of the light states
// ------------------------------------------------------------------------------------------------------------------------------------------------>  
void DumpLightSchemeToSerial(int WhatScheme)
{
    int i;
    int j;
    int MySetting;

    Serial.print("SCHEME: ");
    Serial.print(WhatScheme+1);
    Serial.println();
    for (i=0; i<NumLights; i++)
    {
        Serial.print("Light #");
        Serial.print(i+1);
        Serial.print(" States: ");
        for (j=0; j<NumStates; j++)
        {
            MySetting = pgm_read_word_near(&(Schemes[WhatScheme][i][j]));    
            Serial.print(MySetting, DEC);
            Serial.print("  ");
        }
        Serial.println();
    }
    return;
}



// ------------------------------------------------------------------------------------------------------------------------------------------------>  
// SETLIGHT - This sets an individual light to a specific setting
// ------------------------------------------------------------------------------------------------------------------------------------------------>  
void SetLight(int WhatLight, int WhatSetting)
{ 
    switch (WhatSetting)
    {
        case ON:
            TurnOnLight(WhatLight);
            break;
        case OFF:
            TurnOffLight(WhatLight);
            break;
        case BLINK:
            BlinkLight(WhatLight);
            break;
        case FASTBLINK:
            FastBlinkLight(WhatLight);
            break;
        case DIM:
            DimLight(WhatLight);
            break;
    }
}



// ------------------------------------------------------------------------------------------------------------------------------------------------>  
// TURNONLIGHT - Turns a light on
// ------------------------------------------------------------------------------------------------------------------------------------------------>  
void TurnOnLight(int WhatLight)
{
    digitalWrite(LightPin[WhatLight], HIGH);
}



// ------------------------------------------------------------------------------------------------------------------------------------------------>  
// TURNOFFLIGHT - Turns a light off
// ------------------------------------------------------------------------------------------------------------------------------------------------>  
void TurnOffLight(int WhatLight)
{
    digitalWrite(LightPin[WhatLight], LOW);
}



// ------------------------------------------------------------------------------------------------------------------------------------------------>  
// BLINKLIGHT - This blinks a light
// ------------------------------------------------------------------------------------------------------------------------------------------------>  
void BlinkLight(int WhatLight)
{
    Blinker ? TurnOnLight(WhatLight) : TurnOffLight(WhatLight);
}



// ------------------------------------------------------------------------------------------------------------------------------------------------>  
// FASTBLINKLIGHT - This blinks a light at the fast rate
// ------------------------------------------------------------------------------------------------------------------------------------------------>  
void FastBlinkLight(int WhatLight)
{
    FastBlinker ? TurnOnLight(WhatLight) : TurnOffLight(WhatLight);
}



// ------------------------------------------------------------------------------------------------------------------------------------------------>  
// DIMLIGHT - This dims a light
// ------------------------------------------------------------------------------------------------------------------------------------------------>  
void DimLight(int WhatLight)
{
    // Only certain lights can be dimmed. We do a check here to make sure. 
    // If dimming is not an option on this pin, but the command was still issued, 
    // we instead turn it off. 
    
    // NOTE: Only use the ActualDimLevel variable in this function, do not use DimLevel from UserConfig.h
    if (Dimmable[WhatLight])
    {
        analogWrite(LightPin[WhatLight], ActualDimLevel);
    }
    else
    {
        digitalWrite(LightPin[WhatLight], LOW);
    }
}



// ------------------------------------------------------------------------------------------------------------------------------------------------>  
// FIXDIMLEVEL - Sets the ActualDimLevel and prevents the bug that occurs if the User strangely set DimLevel to less than 2 (unlikely)
// ------------------------------------------------------------------------------------------------------------------------------------------------>  
void FixDimLevel()
{
    if (DimLevel < 2) 
    { ActualDimLevel = 2; }
    else
    { ActualDimLevel = DimLevel; }
}



// ------------------------------------------------------------------------------------------------------------------------------------------------>  
// TWINKLELIGHTS - Sets all lights to FastBlink for Seconds seconds
// ------------------------------------------------------------------------------------------------------------------------------------------------>  
void TwinkleLights(int Seconds)
{
    int j;
    StartWaiting_sec(Seconds);
    do
    {
        for (j=0; j<NumLights; j++)
        {    
            FastBlinkLight(j);
        }
        timer.run();
    } 
    while(!TimeUp);
}
   


// ------------------------------------------------------------------------------------------------------------------------------------------------>  
// BLINKALLLIGHTS - Blinks every single light, onboard and off, HowManyTimes. This is used as an indicator during Change-Scheme-Mode
// ------------------------------------------------------------------------------------------------------------------------------------------------>  
void BlinkAllLights(int HowManyTimes)
{
    if (Blinking)
    {
        if (State)
        {    
            if (!PriorState)
            {    
                // Turn everything on
                GreenLedOn();
                RedLedOn();
                for (int j=0; j<NumLights; j++)
                {   TurnOnLight(j); } 
                PriorState = true;
                TimesBlinked += 1;
                timer.setTimeout(130, BlinkOn);
            }
        }
        else
        {
            if (PriorState)
            {
                // Turn everything off
                GreenLedOff();
                RedLedOff();
                for (int j=0; j<NumLights; j++)
                {   TurnOffLight(j); }        
                PriorState = false;
                BlinkOffID = timer.setTimeout(160, BlinkOff);
            }

            if (TimesBlinked >= HowManyTimes)
            {   TimesBlinked = 0;
                Blinking = false;
                timer.setTimeout(1000, BlinkWait);
                timer.deleteTimer(BlinkOffID);
            }

        }
    }
}

void BlinkOn()
{
    Blinking = true;
    State = false;
    PriorState = true;
}

void BlinkOff()
{
    Blinking = true;
    State = true;
    PriorState = false;
}

void BlinkWait()
{
    Blinking = true;
    State = true;
    PriorState = false;
}
